local addonName = ...
local frameName = "BattlePetTabs"
local watcher, numTabs, numPets, numPets2, elapsed = CreateFrame("Frame"), 0, 0, 0, 0
BattlePetTabsDB = {} -- populated later

local MAX_ACTIVE_PETS = 3
local MAX_ACTIVE_ABILITIES = 3
local MAX_PET_LEVEL = 25

local petTypePassiveChart = {
  [1] = {4, 5},  -- Humanoid    +Undead      -Critter
  [2] = {1, 3},  -- Dragon      +Humanoid    -Flying
  [3] = {6, 8},  -- Flying      +Magical     -Beast
  [4] = {5, 2},  -- Undead      +Critter     -Dragon
  [5] = {8, 1},  -- Critter     +Beast       -Humanoid
  [6] = {2, 9},  -- Magical     +Dragon      -Water
  [7] = {9, 10}, -- Elemental   +Water       -Mechanical
  [8] = {10, 1}, -- Beast       +Mechanical  -Humanoid
  [9] = {3, 4},  -- Water       +Flying      -Undead
  [10] = {7, 6}, -- Mechanical  +Elemental   -Magical
}

local function GetStatIconString(i)
  return "|TInterface\\PetBattles\\BattleBar-AbilityBadge-"..(i and "Strong" or "Weak")..":18:18:0:-2|t" -- 18:18 looks good
end

local function GetPetIconString(i, s) -- needs more work, the dimensions get weird when the image is too big and it creates too much padding
  return "|TInterface\\PetBattles\\PetIcon-"..(type(i) == "string" and i or (type(i) == "number" and PET_TYPE_SUFFIX[i] or "NO_TYPE"))..":22:22:3:"..(s and "2" or "4.25")..":128:256:62:128:102:168|t" -- 18:18 too small, 22:22 is better
end

local function BuildPetTooltipString(petId)
  local speciesId, customName, level, xp, maxXp, displayID, petName, petIcon, petType, creatureID = C_PetJournal.GetPetInfoByPetID(petId)
  if not speciesId then
    speciesId, petName, petIcon = 0, "Unknown", "Interface\\Icons\\INV_Misc_QuestionMark"
  elseif customName then
    petName = customName
  end
  return GetPetIconString(select(3, C_PetJournal.GetPetInfoBySpeciesID(speciesId)), nil).."|T"..petIcon..":18:18|t |cffCCCCCCL"..level.."|r "..petName..(maxXp > 0 and level < MAX_PET_LEVEL and " |cffCCCCCC("..math.floor(xp/maxXp*100).."% exp)|r" or "") -- 0:0 too small, 18:18 is better
end

local function BuildStrongWeakTooltip(petIds, isAttack, isWeak)
  local speciesId, abilityId, petType
  local modifier, matchStrong, matchWeak
  local temp = {strong = {}, weak = {}, sKeys = {}, wKeys = {}}
  for _, petId in pairs(petIds) do
    speciesId = C_PetJournal.GetPetInfoByPetID(petId)
    if speciesId then
      abilityId = PET_BATTLE_PET_TYPE_PASSIVES[select(3, C_PetJournal.GetPetInfoBySpeciesID(speciesId))]
      petType = select(7, C_PetBattles.GetAbilityInfoByID(abilityId))
      if isAttack then
        if petType then
          for i = 1, C_PetJournal.GetNumPetTypes() do
            modifier = C_PetBattles.GetAttackModifier(petType, i)
            if modifier > 1 then
              temp.strong[PET_TYPE_SUFFIX[i]] = GetPetIconString(i, 1)
              table.insert(temp.sKeys, PET_TYPE_SUFFIX[i])
            elseif modifier < 1 then
              temp.weak[PET_TYPE_SUFFIX[i]] = GetPetIconString(i, 1)
              table.insert(temp.wKeys, PET_TYPE_SUFFIX[i])
            end
          end
        end
      else
        if petType then
          matchStrong, matchWeak = unpack(petTypePassiveChart[petType] or {})
          if matchStrong then
            temp.strong[matchStrong] = GetPetIconString(matchStrong, 1)
            table.insert(temp.sKeys, matchStrong)
          end
          if matchWeak then
            temp.weak[matchWeak] = GetPetIconString(matchWeak, 1)
            table.insert(temp.wKeys, matchWeak)
          end
        end
      end
    end
  end
  local lookup = isWeak and temp.wKeys or temp.sKeys
  local lookup2 = isWeak and temp.weak or temp.strong
  table.sort(lookup)
  local text = ""
  for _, key in ipairs(lookup) do
    text = text..lookup2[key]
  end
  return text:trim()
end

local function ApplyRename(teamId, newName)
  if type(newName) == "string" and newName:len() > 0 then
    teamId = tonumber(teamId) or 0
    if teamId > 0 and teamId <= numTabs then
      BattlePetTabsDB[teamId].name = newName
      GameTooltip:Hide()
      BattlePetTabs_Update()
    end
  end
end

local function ApplyDelete(teamId)
  teamId = tonumber(teamId) or 0
  if teamId > 0 and teamId <= numTabs then
    table.wipe(BattlePetTabsDB[teamId])
    GameTooltip:Hide()
    BattlePetTabsReorganizeDB()
    BattlePetTab_LoadTeamId(BattlePetTabsDB.currentId)
    BattlePetTabs_Update()
  end
end

local function GetTeamMacroName(teamId)
  teamId = tonumber(teamId)
  if teamId then
    return "BattlePetTeam"..teamId
  end
end

local function CreateTeamMacro(macroName, teamId, teamData)
  if type(teamData) ~= "table" then
    return
  end
  if GetNumMacros() >= (MAX_ACCOUNT_MACROS or 36) then -- it's a LOD addon but I don't care to add a bunch of code just for one variable, so I hardcoded it!
    print("Can't create macro for team #"..teamId.." because you don't have any more available macro slots in your General category.")
    return
  end
  if MacroFrame then
    HideUIPanel(MacroFrame)
  end
  local iconFile = teamData.icon:lower():gsub(".-\\.-\\(.-)%.blp", "%1") -- need only the filename, the API is strict and doesn't anyway let you define paths outside the Interface\Icons folder
  local macroBody = "#showtooltip\n/run PetJournal_LoadUI()if GetMouseButtonClicked()==\"RightButton\"then TogglePetJournal(2)else local i=tonumber("..(teamId or 0)..")local b=_G[\""..frameName.."Tab\"..i..\"Button\"]if b then b:Click()else print(\""..frameName.." is not loaded.\")end end"
  if GetMacroIndexByName(macroName) == 0 then
    CreateMacro(macroName, 0, macroBody, nil)
  end
  return EditMacro(GetMacroIndexByName(macroName), macroName, iconFile, macroBody)
end

local function ValidateTeamData(petIds)
  if type(petIds) == "table" then
    local temp = {}
    for _, petId in ipairs(petIds) do
      if C_PetJournal.GetPetInfoByPetID(petId) then
        table.insert(temp, petId)
      end
    end
    return next(temp) and temp or nil
  end
end

local function Watcher_OnUpdate(self, elapse)
  elapsed = elapsed + elapse
  if elapsed > .25 then
    elapsed = 0
    if type(BattlePetTabsDB) ~= "table" or type(BattlePetTabsDB.currentId) ~= "number" then
      self:SetScript("OnUpdate", nil)
    else
      BattlePetTab_LoadTeamId(BattlePetTabsDB.currentId)
    end
  end
end

function BattlePetTabsReorganizeDB()
  if type(BattlePetTabsDB) ~= "table" then
    return -- only runs when this is a table
  end
  local temp, lastId, data = {}, 0
  for i = 1, numTabs do
    if type(BattlePetTabsDB[i]) == "table" then
      if #(BattlePetTabsDB[i].team or {}) > 0 then
        data = BattlePetTabsDB[i]
        data.team = ValidateTeamData(data.team)
        table.insert(temp, data)
        lastId = lastId + 1
      end
    end
  end
  for i = lastId + 1, numTabs do
    if type(temp[i]) == "table" then
      table.wipe(temp[i])
    else
      temp[i] = {}
    end
  end
  if type(BattlePetTabsDB.currentId) ~= "number" or BattlePetTabsDB.currentId > lastId then
    temp.currentId = lastId
  else
    temp.currentId = BattlePetTabsDB.currentId
  end
  BattlePetTabsDB = temp
  -- fix orphan macros and update existing teams
  local macroName, macroIndex
  for i = 1, numTabs do
    macroName = GetTeamMacroName(i)
    if macroName then
      if i <= lastId then
        macroIndex = GetMacroIndexByName(macroName)
        if macroIndex ~= 0 then
          CreateTeamMacro(macroName, i, BattlePetTabsDB[i])
        end
      else
        DeleteMacro(macroName)
      end
    end
  end
end

function BattlePetTabs_OnLoad(self)
  numTabs = _G[frameName]:GetNumChildren()
  self:Hide()
  self:RegisterEvent("ADDON_LOADED")
end

function BattlePetTabs_OnEvent(self, event, name)
  if (event == "ADDON_LOADED" and name == addonName) or event == addonName then
    if event == "ADDON_LOADED" then
      self:Show()
      self:UnregisterEvent(event)
      self:RegisterEvent("PET_JOURNAL_LIST_UPDATE")
    end
    BattlePetTabsReorganizeDB()
    if event == addonName then
      BattlePetTab_SetCurentTeamAsTab(BattlePetTabsDB.currentId)
    else
      BattlePetTab_LoadTeamId(BattlePetTabsDB.currentId)
    end
  elseif event == "PET_JOURNAL_LIST_UPDATE" then
    numPets = select(2, C_PetJournal.GetNumPets(true)) -- wild pets only
    if numPets ~= numPets2 then
      numPets2 = numPets
      BattlePetTabsReorganizeDB()
    end
  end
  if type(BattlePetTabsDB) == "table" and type(BattlePetTabsDB.currentId) == "number" then
    BattlePetTabs_Update()
  end
end

function BattlePetTabs_Update()
  local tabData, tabTab, tabButton, tabTexture, shownNewTeam
  for i = 1, numTabs do
    if type(BattlePetTabsDB[i]) ~= "table" then
      BattlePetTabsDB[i] = {}
    end
    tabData = BattlePetTabsDB[i]
    tabTab = _G[frameName.."Tab"..i]
    tabButton = _G[frameName.."Tab"..i.."Button"]
    if shownNewTeam then
      tabTab:Hide()
    else
      tabTexture = _G[frameName.."Tab"..i.."ButtonIconTexture"]
      if not tabData.team then
        tabTexture:SetTexture("Interface\\GuildBankFrame\\UI-GuildBankFrame-NewTab")
        tabButton.tooltip = "New Team"
        tabButton.newTeam = 1
        shownNewTeam = 1
      else
        tabTexture:SetTexture(tabData.icon)
        tabButton.tooltip = {}
        table.insert(tabButton.tooltip, tabData.name)
        tabButton.newTeam = nil
        for _, petId in ipairs(tabData.team) do
          table.insert(tabButton.tooltip, BuildPetTooltipString(petId))
        end
        table.insert(tabButton.tooltip, "Atk."..GetStatIconString(1).."vs: "..BuildStrongWeakTooltip(tabData.team, 1))
        table.insert(tabButton.tooltip, "Atk."..GetStatIconString().."vs: "..BuildStrongWeakTooltip(tabData.team, 1, 1))
        table.insert(tabButton.tooltip, "Def."..GetStatIconString(1).."vs: "..BuildStrongWeakTooltip(tabData.team, nil, 1))
        table.insert(tabButton.tooltip, "Def."..GetStatIconString().."vs: "..BuildStrongWeakTooltip(tabData.team))
      end
      if i == BattlePetTabsDB.currentId then
        tabButton:SetChecked(1)
      else
        tabButton:SetChecked(nil)
      end
      tabTab:Show()
    end
  end
  local focus = GetMouseFocus() -- the tooltip stays the same when we add or remove teams, so this helps update that tooltip automatically by checking what we are currently hovering over and triggering the mechanism is appropriate
  if type(focus) == "table" and type(focus.GetObjectType) == "function" and focus:GetObjectType() == "CheckButton" and type(focus.GetScript) == "function" then
    (focus:GetScript("OnEnter") or function()end)(focus)
  end
end

function BattlePetTab_SetCurentTeamAsTab(teamId)
  if type(BattlePetTabsDB[teamId]) ~= "table" then
    BattlePetTabsDB[teamId] = {}
  end
  local team, temp, ab1, ab2, ab3 = BattlePetTabsDB[teamId]
  if type(team.name) ~= "string" then
    team.name = "Team "..teamId
  end
  team.team = {}
  team.team2 = {}
  for i = 1, MAX_ACTIVE_PETS do
    temp, ab1, ab2, ab3 = C_PetJournal.GetPetLoadOutInfo(i)
    if temp > 0 then
      table.insert(team.team, temp)
      team.team2[i] = {ab1, ab2, ab3}
    end
  end
  team.icon = nil -- refresh icon
  for _, petId in ipairs(team.team) do
    team.icon = select(8, C_PetJournal.GetPetInfoByPetID(petId))
    if team.icon then
      break
    end
  end
  if not team.icon then
    team.icon = "Interface\\Icons\\INV_Misc_QuestionMark"
  end
  BattlePetTabsReorganizeDB()
end

function BattlePetTab_LoadTeamId(teamId)
  if type(BattlePetTabsDB) == "table" and type(BattlePetTabsDB[teamId]) == "table" and type(BattlePetTabsDB[teamId].team) == "table" then
    local pets, count, unhook, abilities = {}, 0, nil, nil
    for i = 1, MAX_ACTIVE_PETS do
      pets[i] = BattlePetTabsDB[teamId].team[i]
      pets[i] = type(pets[i]) == "number" and pets[i] or 0
      pets[i] = pets[i] <= 0 and 0 or pets[i]
      if pets[i] == C_PetJournal.GetPetLoadOutInfo(i) then
        count = count + 1
      else
        C_PetJournal.SetPetLoadOutInfo(i, pets[i], addonName)
      end
    end
    if count == 3 then
      unhook = 1
      if type(BattlePetTabsDB[teamId].team2) == "table" then -- legacy support (when team doesn't have abilities table, this avoids lua errors)
        abilities = BattlePetTabsDB[teamId].team2
        count = 0
        for i, petId in ipairs(pets) do
          if type(abilities[i]) == "table" then
            for j = 1, MAX_ACTIVE_ABILITIES do
              if abilities[i][j] == select(j + 1, C_PetJournal.GetPetLoadOutInfo(i)) then
                count = count + 1
              else
                C_PetJournal.SetAbility(i, j, abilities[i][j], addonName)
              end
            end
          else
            count = count + MAX_ACTIVE_ABILITIES -- trick it to think we equipped the spells we need (since this bugged it means there are no spells in the cache anyway, so we need to make the player re-pick their choices and hopefully this will fix the bug, if not we need to debug some more!)
          end
        end
        if count ~= MAX_ACTIVE_PETS * MAX_ACTIVE_ABILITIES then
          unhook = nil
        end
      end
      if unhook then
        watcher:SetScript("OnUpdate", nil)
        if PetJournal_UpdatePetLoadOut then
          PetJournal_UpdatePetLoadOut()
        elseif PetJournal_UpdateAll then
          PetJournal_UpdateAll()
        else
          print("Fatal error, couldn't update Pet Journal UI!")
        end
      else
        watcher:SetScript("OnUpdate", Watcher_OnUpdate)
      end
    else
      watcher:SetScript("OnUpdate", Watcher_OnUpdate)
    end
  end
end

function BattlePetTab_OnClick(self, button, currentId)
  if button == "LeftButton" then
    StaticPopup_Hide("BATTLETABS_TEAM_RENAME")
    StaticPopup_Hide("BATTLETABS_TEAM_DELETE")
  end
  if button ~= "LeftButton" and self.newTeam then
  else
    BattlePetTabsDB.currentId = currentId
    if button == "LeftButton" then
      if self.newTeam then
        BattlePetTab_SetCurentTeamAsTab(currentId, 1)
      else
        BattlePetTab_LoadTeamId(currentId)
      end
    elseif button == "RightButton" and (not self.newTeam) then
      BattlePetTab_LoadTeamId(currentId)
      if IsModifiedClick() then
        StaticPopup_Show("BATTLETABS_TEAM_RENAME", BattlePetTabsDB[currentId].name, nil, currentId)
      else
        StaticPopup_Show("BATTLETABS_TEAM_DELETE", BattlePetTabsDB[currentId].name, nil, currentId)
      end
    end
  end
  BattlePetTabs_Update()
end

function BattlePetTab_OnDrag(self, button, currentId)
  if (not InCombatLockdown()) and (not self.newTeam) then -- can't work with macros in combat
    local macroName = GetTeamMacroName(currentId)
    if macroName then
      local macroId = CreateTeamMacro(macroName, currentId, BattlePetTabsDB[currentId])
      if macroId then
        ClearCursor()
        PickupMacro(macroId)
      end
    end
  end
end

local function UpdateCurrentTeamLoadOut(slotId, petId, skipUpdating, ...)
  if skipUpdating ~= addonName then
    BattlePetTabs_OnEvent(BattlePetTabs, addonName)
  end
end

local function UpdateCurrentTeamLoadOut2(slotId, abilitySlot, abilityId, skipUpdating, ...)
  if skipUpdating ~= addonName then
    BattlePetTabs_OnEvent(BattlePetTabs, addonName)
  end
end

local function onGameTooltipShow(self)
  local text, tabButton = _G[self:GetName().."TextLeft1"]:GetText()
  for i = 1, numTabs do
    if text and text == GetTeamMacroName(i) then
      tabButton = _G[frameName.."Tab"..i.."Button"]
      if tabButton and tabButton.tooltip then
        self:ClearLines()
        if type(tabButton.tooltip) == "table" then
          for _, line in ipairs(tabButton.tooltip) do
            self:AddLine(line)
          end
        else
          self:AddLine(tabButton.tooltip)
        end
        self:AddLine("|cff999999Right-click to open the Pet Journal|r")
        self:Show()
      end
      break
    end
  end
end

hooksecurefunc(C_PetJournal, "SetPetLoadOutInfo", UpdateCurrentTeamLoadOut)
hooksecurefunc(C_PetJournal, "SetAbility", UpdateCurrentTeamLoadOut2)
GameTooltip:HookScript("OnShow", onGameTooltipShow)

StaticPopupDialogs["BATTLETABS_TEAM_RENAME"] = {
  text = "What do you wish to rename |cffffd200%s|r to?",
  button1 = ACCEPT,
  button2 = CANCEL,
  hasEditBox = 1,
  maxLetters = 30,
  OnAccept = function(self)
    ApplyRename(self.data, self.editBox:GetText())
  end,
  EditBoxOnEnterPressed = function(self)
    local parent = self:GetParent()
    ApplyRename(parent.data, parent.editBox:GetText())
    parent:Hide()
  end,
  OnShow = function(self)
    self.editBox:SetFocus()
  end,
  OnHide = function(self)
    ChatEdit_FocusActiveWindow()
    self.editBox:SetText("")
  end,
  timeout = 0,
  exclusive = 1,
  hideOnEscape = 1,
}

StaticPopupDialogs["BATTLETABS_TEAM_DELETE"] = {
	text = "Are you sure you want to delete team |cffffd200%s|r?",
	button1 = OKAY,
	button2 = CANCEL,
	OnAccept = function(self)
    ApplyDelete(self.data)
	end,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1,
}
